import hypothesis.strategies as st

from .test_connection import TestConnection as _TestConnection
from .test_server import TestServer as _TestServer
from .test_zset import TestZSet as _TestZSet
from .test_set import TestSet as _TestSet
from .test_list import TestList as _TestList
from .test_hash import TestHash as _TestHash
from .base import BaseTest, common_commands, commands
from .test_string import TestString as _TestString, string_commands

bad_commands = (
    # redis-py splits the command on spaces, and hangs if that ends up being an empty list
    commands(st.text().filter(lambda x: bool(x.split())), st.lists(st.binary() | st.text()))
)


class TestJoint(BaseTest):
    create_command_strategy = (
        _TestString.create_command_strategy
        | _TestHash.create_command_strategy
        | _TestList.create_command_strategy
        | _TestSet.create_command_strategy
        | _TestZSet.create_command_strategy
    )
    command_strategy = (
        _TestServer.server_commands
        | _TestConnection.connection_commands
        | string_commands
        | _TestHash.hash_commands
        | _TestList.list_commands
        | _TestSet.set_commands
        | _TestZSet.zset_commands
        | common_commands
        | bad_commands
    )



import time
from datetime import datetime, timedelta

import pytest
import redis
from redis import exceptions

pytest.importorskip("jsonpath_ng")
try:
    from redis.commands.core import DataPersistOptions
except ImportError:
    pytest.skip()

from test.testtools import raw_command

pytestmark = []
pytestmark.extend(
    [
        pytest.mark.min_server("8.4"),
    ]
)


def redis_server_time(client):
    seconds, milliseconds = client.time()
    timestamp = float(f"{seconds}.{milliseconds}")
    return datetime.fromtimestamp(timestamp)


def test_msetex_no_expiration_with_cluster_client(r: redis.Redis):
    all_test_keys = ["1:{test:1}", "2:{test:1}"]
    for key in all_test_keys:
        r.delete(key)

    # set items from mapping without expiration
    assert r.msetex(mapping={"1:{test:1}": 1, "2:{test:1}": b"four"}) == 1
    assert r.mget("1:{test:1}", "2:{test:1}") == [b"1", b"four"]
    assert r.ttl("1:{test:1}") == -1
    assert r.ttl("2:{test:1}") == -1


def test_msetex_expiration_ex_and_keepttl_with_cluster_client(r: redis.Redis):
    all_test_keys = ["1:{test:1}", "2:{test:1}"]
    for key in all_test_keys:
        r.delete(key)

    # set items from mapping with expiration - testing ex field
    assert (
        r.msetex(
            mapping={"1:{test:1}": 1, "2:{test:1}": "2"},
            ex=10,
        )
        == 1
    )
    ttls = [r.ttl(key) for key in all_test_keys]
    for ttl in ttls:
        assert pytest.approx(ttl) == 10

    assert r.mget(*all_test_keys) == [b"1", b"2"]
    time.sleep(1.1)
    # validate keepttl
    assert r.msetex(mapping={"1:{test:1}": 11}, keepttl=True) == 1
    assert r.ttl("1:{test:1}") < 10


def test_msetex_expiration_px_with_cluster_client(r: redis.Redis):
    all_test_keys = ["1:{test:1}", "2:{test:1}"]
    for key in all_test_keys:
        r.delete(key)

    mapping = {"1:{test:1}": 1, "2:{test:1}": "2"}
    # set key/value pairs provided in mapping
    # with expiration - testing px field
    assert r.msetex(mapping=mapping, px=60000) == 1

    ttls = [r.ttl(key) for key in mapping.keys()]
    for ttl in ttls:
        assert pytest.approx(ttl) == 60
    assert r.mget(*mapping.keys()) == [b"1", b"2"]


def test_msetex_expiration_pxat_and_nx_with_cluster_client(r: redis.Redis):
    all_test_keys = [
        "1:{test:1}",
        "2:{test:1}",
        "3:{test:1}",
        "new:{test:1}",
        "new_2:{test:1}",
    ]
    for key in all_test_keys:
        r.delete(key)

    mapping = {"1:{test:1}": 1, "2:{test:1}": "2", "3:{test:1}": "three"}
    assert r.msetex(mapping=mapping, ex=30) == 1

    # NX is set with existing keys - nothing should be saved or updated
    expire_at = redis_server_time(r) + timedelta(seconds=10)
    assert (
        r.msetex(
            mapping={"1:{test:1}": "new_value", "new:{test:1}": "ok"},
            pxat=expire_at,
            data_persist_option=DataPersistOptions.NX,
        )
        == 0
    )

    ttls = [r.ttl(key) for key in mapping.keys()]
    for ttl in ttls:
        assert 10 < ttl <= 30
    assert r.mget(*mapping.keys(), "new:{test:1}") == [b"1", b"2", b"three", None]

    # NX is set with non existing keys - values should be set
    assert (
        r.msetex(
            mapping={"new:{test:1}": "ok", "new_2:{test:1}": "ok_2"},
            pxat=expire_at,
            data_persist_option=DataPersistOptions.NX,
        )
        == 1
    )
    old_ttls = [r.ttl(key) for key in mapping.keys()]
    new_ttls = [r.ttl(key) for key in ["new:{test:1}", "new_2:{test:1}"]]
    for ttl in old_ttls:
        assert 10 < ttl <= 30
    for ttl in new_ttls:
        assert ttl <= 11
    assert r.mget(*mapping.keys(), "new:{test:1}", "new_2:{test:1}") == [
        b"1",
        b"2",
        b"three",
        b"ok",
        b"ok_2",
    ]


def test_msetex_expiration_exat_and_xx_with_cluster_client(r: redis.Redis):
    all_test_keys = ["1:{test:1}", "2:{test:1}", "3:{test:1}", "new:{test:1}"]
    for key in all_test_keys:
        r.delete(key)

    mapping = {"1:{test:1}": 1, "2:{test:1}": "2", "3:{test:1}": "three"}
    assert r.msetex(mapping, ex=30) == 1

    expire_at = redis_server_time(r) + timedelta(seconds=10)
    # XX is set with unexisting key - nothing should be saved or updated
    assert (
        r.msetex(
            mapping={"1:{test:1}": "new_value", "new:{test:1}": "ok"},
            exat=expire_at,
            data_persist_option=DataPersistOptions.XX,
        )
        == 0
    )
    ttls = [r.ttl(key) for key in mapping.keys()]
    for ttl in ttls:
        assert 10 < ttl <= 30
    assert r.mget(*mapping.keys(), "new:{test:1}") == [b"1", b"2", b"three", None]

    # XX is set with existing keys - values should be updated
    assert (
        r.msetex(
            mapping={"1:{test:1}": "new_value", "2:{test:1}": "new_value_2"},
            exat=expire_at,
            data_persist_option=DataPersistOptions.XX,
        )
        == 1
    )
    ttls = [r.ttl(key) for key in mapping.keys()]
    assert ttls[0] <= 11
    assert ttls[1] <= 11
    assert 10 < ttls[2] <= 30
    assert r.mget("1:{test:1}", "2:{test:1}", "3:{test:1}", "new:{test:1}") == [
        b"new_value",
        b"new_value_2",
        b"three",
        None,
    ]


def test_msetex_no_expiration(r: redis.Redis):
    all_test_keys = ["1", "2"]
    for key in all_test_keys:
        r.delete(key)

    # # set items from mapping without expiration
    assert r.msetex(mapping={"1": 1, "2": b"four"}) == 1
    assert r.mget("1", "2") == [b"1", b"four"]
    assert r.ttl("1") == -1
    assert r.ttl("2") == -1


def test_msetex_expiration_ex_and_keepttl(r: redis.Redis):
    all_test_keys = ["1", "2"]
    for key in all_test_keys:
        r.delete(key)

    # set items from mapping with expiration - testing ex field
    assert (
        r.msetex(
            mapping={"1": 1, "2": "2"},
            ex=10,
        )
        == 1
    )
    ttls = [r.ttl(key) for key in all_test_keys]
    for ttl in ttls:
        assert pytest.approx(ttl) == 10

    assert r.mget(*all_test_keys) == [b"1", b"2"]
    time.sleep(1.1)
    # validate keepttl
    assert r.msetex(mapping={"1": 11}, keepttl=True) == 1
    assert r.ttl("1") < 10


def test_msetex_expiration_px(r: redis.Redis):
    all_test_keys = ["1", "2"]
    for key in all_test_keys:
        r.delete(key)

    mapping = {"1": 1, "2": "2"}
    # set key/value pairs provided in mapping
    # with expiration - testing px field
    assert r.msetex(mapping=mapping, px=60000) == 1

    ttls = [r.ttl(key) for key in mapping.keys()]
    for ttl in ttls:
        assert pytest.approx(ttl) == 60
    assert r.mget(*mapping.keys()) == [b"1", b"2"]


def test_msetex_expiration_pxat_and_nx(r: redis.Redis):
    all_test_keys = ["1", "2", "3", "new", "new_2"]
    for key in all_test_keys:
        r.delete(key)

    mapping = {"1": 1, "2": "2", "3": "three"}
    assert r.msetex(mapping=mapping, ex=30) == 1

    # NX is set with existing keys - nothing should be saved or updated
    expire_at = redis_server_time(r) + timedelta(seconds=20)
    assert (
        r.msetex(
            mapping={"1": "new_value", "new": "ok"},
            pxat=expire_at,
            data_persist_option=DataPersistOptions.NX,
        )
        == 0
    )
    ttls = [r.ttl(key) for key in mapping.keys()]
    for ttl in ttls:
        assert 10 < ttl <= 30
    assert r.mget(*mapping.keys(), "new") == [b"1", b"2", b"three", None]

    # NX is set with non existing keys - values should be set
    assert (
        r.msetex(
            mapping={"new": "ok", "new_2": "ok_2"},
            pxat=expire_at,
            data_persist_option=DataPersistOptions.NX,
        )
        == 1
    )
    old_ttls = [r.ttl(key) for key in mapping.keys()]
    new_ttls = [r.ttl(key) for key in ["new", "new_2"]]
    for ttl in old_ttls:
        assert 10 < ttl <= 30
    for ttl in new_ttls:
        assert ttl <= 21
    assert r.mget(*mapping.keys(), "new", "new_2") == [
        b"1",
        b"2",
        b"three",
        b"ok",
        b"ok_2",
    ]


def test_msetex_expiration_exat_and_xx(r: redis.Redis):
    all_test_keys = ["1", "2", "3", "new"]
    for key in all_test_keys:
        r.delete(key)

    mapping = {"1": 1, "2": "2", "3": "three"}
    assert r.msetex(mapping, ex=30) == 1

    expire_at = redis_server_time(r) + timedelta(seconds=10)
    ## XX is set with nonexistent key - nothing should be saved or updated
    assert (
        r.msetex(
            mapping={"1": "new_value", "new": "ok"},
            exat=expire_at,
            data_persist_option=DataPersistOptions.XX,
        )
        == 0
    )
    ttls = [r.ttl(key) for key in mapping.keys()]
    for ttl in ttls:
        assert 10 < ttl <= 30
    assert r.mget(*mapping.keys(), "new") == [b"1", b"2", b"three", None]

    # XX is set with existing keys - values should be updated
    assert (
        r.msetex(
            mapping={"1": "new_value", "2": "new_value_2"},
            exat=expire_at,
            data_persist_option=DataPersistOptions.XX,
        )
        == 1
    )
    ttls = [r.ttl(key) for key in mapping.keys()]
    assert ttls[0] <= 11
    assert ttls[1] <= 11
    assert 10 < ttls[2] <= 30
    assert r.mget("1", "2", "3", "new") == [
        b"new_value",
        b"new_value_2",
        b"three",
        None,
    ]


def test_msetex_invalid_ex_and_keepttl(r: redis.Redis):
    with pytest.raises(exceptions.ResponseError) as excinfo:
        raw_command(r, "MSETEX", 2, "1", 1, "2", 2, "EX", 10, "KEEPTTL")
    assert str(excinfo.value) == "syntax error"


def test_msetex_invalid_ex(r: redis.Redis):
    with pytest.raises(exceptions.ResponseError) as excinfo:
        raw_command(r, "MSETEX", 2, "1", 1, "2", 2, "EX", -1)
    assert str(excinfo.value) == "invalid expire time in 'msetex' command"


def test_msetex_invalid_nx_and_xx(r: redis.Redis):
    with pytest.raises(exceptions.ResponseError) as excinfo:
        raw_command(r, "MSETEX", 2, "1", 1, "2", 2, "EX", 10, "XX", "NX")
    assert str(excinfo.value) == "syntax error"


def test_msetex_invalid_wrong_number_of_keys(r: redis.Redis):
    with pytest.raises(exceptions.ResponseError) as excinfo:
        raw_command(r, "MSETEX", 1, "1", 1, "2", 2, "EX", 10, "XX", "NX")
    assert str(excinfo.value) == "syntax error"


def test_msetex_invalid_negative_number_of_keys(r: redis.Redis):
    with pytest.raises(exceptions.ResponseError) as excinfo:
        raw_command(r, "MSETEX", -1, "1", 1, "2", 2, "EX", 10, "XX", "NX")
    assert str(excinfo.value) == "invalid numkeys value"


def test_msetex_invalid_less_keys(r: redis.Redis):
    with pytest.raises(exceptions.ResponseError) as excinfo:
        raw_command(r, "MSETEX", 5, "1", 1, "2", 2)
    assert str(excinfo.value) == "wrong number of key-value pairs"



from typing import Tuple

import pytest
import redis
from redis.lock import Lock

pytest.importorskip("lupa")

pytestmark = []
pytestmark.extend(
    [
        pytest.mark.tcp_server,
    ]
)


def test_evalsha_missing_script(tcp_server_address: Tuple[str, int]):
    """Test that EVALSHA with a non-existent script returns NOSCRIPT error."""
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        fake_sha = "0" * 40
        with pytest.raises(redis.exceptions.NoScriptError):
            r.evalsha(fake_sha, 0)


def test_tcp_server_lock(tcp_server_address: Tuple[str, int]):
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        lock = Lock(r, "my-lock")
        lock.acquire()
        assert lock.locked() is True
        lock.release()


def test_eval_multiline_script(tcp_server_address: Tuple[str, int]):
    """Test that EVAL works with multi-line Lua scripts."""
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        # Multi-line script with trailing newline
        script = """
local key = KEYS[1]
local value = ARGV[1]
redis.call('SET', key, value)
return redis.call('GET', key)
"""
        result = r.eval(script, 1, "testkey", "testvalue")
        assert result == b"testvalue"


def test_script_load_multiline(tcp_server_address: Tuple[str, int]):
    """Test that SCRIPT LOAD works with multi-line Lua scripts."""
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        # Multi-line script
        script = """local x = 1
local y = 2
return x + y"""
        sha = r.script_load(script)
        result = r.evalsha(sha, 0)
        assert result == 3


def test_eval_script_with_trailing_newline(tcp_server_address: Tuple[str, int]):
    """Test that scripts with trailing newlines are preserved."""
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        # Script with explicit trailing newline
        script = "return 'hello'\n"
        result = r.eval(script, 0)
        assert result == b"hello"



from __future__ import annotations

import gc
import logging
import weakref
from typing import cast

import pytest
import redis
import redis.client
from redis.exceptions import ResponseError

import fakeredis
from test import testtools
from test.testtools import raw_command

_ = pytest.importorskip("lupa")


@pytest.mark.min_server("7")
def test_script_exists_redis7(r: redis.Redis):
    # test response for no arguments by bypassing the py-redis command
    # as it requires at least one argument
    with pytest.raises(redis.ResponseError):
        raw_command(r, "SCRIPT EXISTS")

    # use single character characters for non-existing scripts, as those
    # will never be equal to an actual sha1 hash digest
    assert r.script_exists("a") == [0]
    assert r.script_exists("a", "b", "c", "d", "e", "f") == [0, 0, 0, 0, 0, 0]

    sha1_one = r.script_load("return 'a'")
    assert r.script_exists(sha1_one) == [1]
    assert r.script_exists(sha1_one, "a") == [1, 0]
    assert r.script_exists("a", "b", "c", sha1_one, "e") == [0, 0, 0, 1, 0]

    sha1_two = r.script_load("return 'b'")
    assert r.script_exists(sha1_one, sha1_two) == [1, 1]
    assert r.script_exists("a", sha1_one, "c", sha1_two, "e", "f") == [0, 1, 0, 1, 0, 0]


@pytest.mark.parametrize("args", [("a",), tuple("abcdefghijklmn")])
def test_script_flush_errors_with_args(r, args):
    with pytest.raises(redis.ResponseError):
        raw_command(r, "SCRIPT FLUSH %s" % " ".join(args))


def test_script_flush(r: redis.Redis):
    # generate/load six unique scripts and store their sha1 hash values
    sha1_values = [r.script_load("return '%s'" % char) for char in "abcdef"]

    # assert the scripts all exist prior to flushing
    assert r.script_exists(*sha1_values) == [1] * len(sha1_values)

    # flush and assert OK response
    assert r.script_flush() is True

    # assert none of the scripts exists after flushing
    assert r.script_exists(*sha1_values) == [0] * len(sha1_values)


def test_script_no_subcommands(r: redis.Redis):
    with pytest.raises(redis.ResponseError):
        raw_command(r, "SCRIPT")


@pytest.mark.max_server("7")
def test_script_help(r: redis.Redis):
    assert raw_command(r, "SCRIPT HELP") == [
        b"SCRIPT <subcommand> [<arg> [value] [opt] ...]. Subcommands are:",
        b"DEBUG (YES|SYNC|NO)",
        b"    Set the debug mode for subsequent scripts executed.",
        b"EXISTS <sha1> [<sha1> ...]",
        b"    Return information about the existence of the scripts in the script cache.",
        b"FLUSH [ASYNC|SYNC]",
        b"    Flush the Lua scripts cache. Very dangerous on replicas.",
        b"    When called without the optional mode argument, the behavior is determined by the",
        b"    lazyfree-lazy-user-flush configuration directive. Valid modes are:",
        b"    * ASYNC: Asynchronously flush the scripts cache.",
        b"    * SYNC: Synchronously flush the scripts cache.",
        b"KILL",
        b"    Kill the currently executing Lua script.",
        b"LOAD <script>",
        b"    Load a script into the scripts cache without executing it.",
        b"HELP",
        b"    Prints this help.",
    ]


@pytest.mark.min_server("7.1")
@pytest.mark.unsupported_server_types("valkey")
def test_script_help73(r: redis.Redis):
    assert raw_command(r, "SCRIPT HELP") == [
        b"SCRIPT <subcommand> [<arg> [value] [opt] ...]. Subcommands are:",
        b"DEBUG (YES|SYNC|NO)",
        b"    Set the debug mode for subsequent scripts executed.",
        b"EXISTS <sha1> [<sha1> ...]",
        b"    Return information about the existence of the scripts in the script cache.",
        b"FLUSH [ASYNC|SYNC]",
        b"    Flush the Lua scripts cache. Very dangerous on replicas.",
        b"    When called without the optional mode argument, the behavior is determined by the",
        b"    lazyfree-lazy-user-flush configuration directive. Valid modes are:",
        b"    * ASYNC: Asynchronously flush the scripts cache.",
        b"    * SYNC: Synchronously flush the scripts cache.",
        b"KILL",
        b"    Kill the currently executing Lua script.",
        b"LOAD <script>",
        b"    Load a script into the scripts cache without executing it.",
        b"HELP",
        b"    Print this help.",
    ]


@pytest.mark.max_server("7.1")
def test_eval_blpop(r: redis.Redis):
    r.rpush("foo", "bar")
    with pytest.raises(redis.ResponseError, match="This Redis command is not allowed from script"):
        r.eval('return redis.pcall("BLPOP", KEYS[1], 1)', 1, "foo")


def test_eval_set_value_to_arg(r: redis.Redis):
    r.eval('redis.call("SET", KEYS[1], ARGV[1])', 1, "foo", "bar")
    val = r.get("foo")
    assert val == b"bar"


def test_eval_conditional(r: redis.Redis):
    lua = """
    local val = redis.call("GET", KEYS[1])
    if val == ARGV[1] then
        redis.call("SET", KEYS[1], ARGV[2])
    else
        redis.call("SET", KEYS[1], ARGV[1])
    end
    """
    r.eval(lua, 1, "foo", "bar", "baz")
    val = r.get("foo")
    assert val == b"bar"
    r.eval(lua, 1, "foo", "bar", "baz")
    val = r.get("foo")
    assert val == b"baz"


def test_eval_table(r: redis.Redis):
    lua = """
    local a = {}
    a[1] = "foo"
    a[2] = "bar"
    a[17] = "baz"
    return a
    """
    val = r.eval(lua, 0)
    assert val == [b"foo", b"bar"]


def test_eval_table_with_nil(r: redis.Redis):
    lua = """
    local a = {}
    a[1] = "foo"
    a[2] = nil
    a[3] = "bar"
    return a
    """
    val = r.eval(lua, 0)
    assert val == [b"foo"]


def test_eval_table_with_numbers(r: redis.Redis):
    lua = """
    local a = {}
    a[1] = 42
    return a
    """
    val = r.eval(lua, 0)
    assert val == [42]


def test_eval_nested_table(r: redis.Redis):
    lua = """
    local a = {}
    a[1] = {}
    a[1][1] = "foo"
    return a
    """
    val = r.eval(lua, 0)
    assert val == [[b"foo"]]


def test_eval_iterate_over_argv(r: redis.Redis):
    lua = """
    for i, v in ipairs(ARGV) do
    end
    return ARGV
    """
    val = r.eval(lua, 0, "a", "b", "c")
    assert val == [b"a", b"b", b"c"]


def test_eval_iterate_over_keys(r: redis.Redis):
    lua = """
    for i, v in ipairs(KEYS) do
    end
    return KEYS
    """
    val = r.eval(lua, 2, "a", "b", "c")
    assert val == [b"a", b"b"]


def test_eval_mget(r: redis.Redis):
    r.set("foo1", "bar1")
    r.set("foo2", "bar2")
    val = r.eval('return redis.call("mget", "foo1", "foo2")', 2, "foo1", "foo2")
    assert val == [b"bar1", b"bar2"]


def test_eval_mget_not_set(r: redis.Redis):
    val = r.eval('return redis.call("mget", "foo1", "foo2")', 2, "foo1", "foo2")
    assert val == [None, None]


def test_eval_hgetall(r: redis.Redis):
    r.hset("foo", "k1", "bar")
    r.hset("foo", "k2", "baz")
    val = r.eval('return redis.call("hgetall", "foo")', 1, "foo")
    sorted_val = sorted([val[:2], val[2:]])
    assert sorted_val == [[b"k1", b"bar"], [b"k2", b"baz"]]


def test_eval_hgetall_iterate(r: redis.Redis):
    r.hset("foo", "k1", "bar")
    r.hset("foo", "k2", "baz")
    lua = """
    local result = redis.call("hgetall", "foo")
    for i, v in ipairs(result) do
    end
    return result
    """
    val = r.eval(lua, 1, "foo")
    sorted_val = sorted([val[:2], val[2:]])
    assert sorted_val == [[b"k1", b"bar"], [b"k2", b"baz"]]


def test_eval_invalid_command(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval('return redis.call("FOO")', 0)


def test_eval_syntax_error(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval('return "', 0)


def test_eval_runtime_error(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval('error("CRASH")', 0)


def test_eval_more_keys_than_args(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval("return 1", 42)


def test_eval_numkeys_float_string(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval("return KEYS[1]", "0.7", "foo")


def test_eval_numkeys_integer_string(r: redis.Redis):
    val = r.eval("return KEYS[1]", "1", "foo")
    assert val == b"foo"


def test_eval_numkeys_negative(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval("return KEYS[1]", -1, "foo")


def test_eval_numkeys_float(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval("return KEYS[1]", 0.7, "foo")


def test_eval_global_variable(r: redis.Redis):
    # Redis doesn't allow script to define global variables
    with pytest.raises(ResponseError):
        r.eval("a=10", 0)


def test_eval_global_and_return_ok(r: redis.Redis):
    # Redis doesn't allow script to define global variables
    with pytest.raises(ResponseError):
        r.eval(
            """
            a=10
            return redis.status_reply("Everything is awesome")
            """,
            0,
        )


def test_eval_convert_number(r: redis.Redis):
    # Redis forces all Lua numbers to integer
    val = r.eval("return 3.2", 0)
    assert val == 3
    val = r.eval("return 3.8", 0)
    assert val == 3
    val = r.eval("return -3.8", 0)
    assert val == -3


def test_eval_convert_bool(r: redis.Redis):
    # Redis converts true to 1 and false to nil (which redis-py converts to None)
    assert r.eval("return false", 0) is None
    val = r.eval("return true", 0)
    assert val == 1
    assert not isinstance(val, bool)


@pytest.mark.min_server("7")
@pytest.mark.unsupported_server_types("valkey")
def test_eval_call_bool7_redis(r: redis.Redis):
    # Redis doesn't allow Lua bools to be passed to [p]call
    with pytest.raises(redis.ResponseError) as exc_info:
        r.eval('return redis.call("SET", KEYS[1], true)', 1, "testkey")
    assert "Lua redis lib command arguments must be strings or integers" in str(exc_info.value)


@pytest.mark.min_server("7")
@pytest.mark.unsupported_server_types("redis")
def test_eval_call_bool7_valkey(r: redis.Redis):
    # Redis doesn't allow Lua bools to be passed to [p]call
    with pytest.raises(redis.ResponseError) as exc_info:
        r.eval('return redis.call("SET", KEYS[1], true)', 1, "testkey")
    assert "Command arguments must be strings or integers script" in str(exc_info.value)


def test_eval_return_error(r: redis.Redis):
    with pytest.raises(redis.ResponseError, match="Testing") as exc_info:
        r.eval('return {err="Testing"}', 0)
    assert isinstance(exc_info.value.args[0], str)
    with pytest.raises(redis.ResponseError, match="Testing") as exc_info:
        r.eval('return redis.error_reply("Testing")', 0)
    assert isinstance(exc_info.value.args[0], str)


def test_eval_return_redis_error(r: redis.Redis):
    with pytest.raises(redis.ResponseError) as exc_info:
        r.eval('return redis.pcall("BADCOMMAND")', 0)
    assert isinstance(exc_info.value.args[0], str)


def test_eval_return_ok(r: redis.Redis):
    val = r.eval('return {ok="Testing"}', 0)
    assert val == b"Testing"
    val = r.eval('return redis.status_reply("Testing")', 0)
    assert val == b"Testing"


def test_eval_return_ok_nested(r: redis.Redis):
    val = r.eval(
        """
        local a = {}
        a[1] = {ok="Testing"}
        return a
        """,
        0,
    )
    assert val == [b"Testing"]


def test_eval_return_ok_wrong_type(r: redis.Redis):
    with pytest.raises(redis.ResponseError):
        r.eval("return redis.status_reply(123)", 0)


def test_eval_pcall(r: redis.Redis):
    val = r.eval(
        """
        local a = {}
        a[1] = redis.pcall("foo")
        return a
        """,
        0,
    )
    assert isinstance(val, list)
    assert len(val) == 1
    assert isinstance(val[0], ResponseError)


def test_eval_pcall_return_value(r: redis.Redis):
    with pytest.raises(ResponseError):
        r.eval('return redis.pcall("foo")', 0)


def test_eval_delete(r: redis.Redis):
    r.set("foo", "bar")
    val = r.get("foo")
    assert val == b"bar"
    val = r.eval('redis.call("DEL", KEYS[1])', 1, "foo")
    assert val is None


def test_eval_exists(r: redis.Redis):
    val = r.eval('return redis.call("exists", KEYS[1]) == 0', 1, "foo")
    assert val == 1


def test_eval_flushdb(r: redis.Redis):
    r.set("foo", "bar")
    val = r.eval(
        """
        local value = redis.call("FLUSHDB");
        return type(value) == "table" and value.ok == "OK";
        """,
        0,
    )
    assert val == 1


def test_eval_flushall(r, create_connection):
    r1 = create_connection(db=2)
    r2 = create_connection(db=3)

    r1["r1"] = "r1"
    r2["r2"] = "r2"

    val = r.eval(
        """
        local value = redis.call("FLUSHALL");
        return type(value) == "table" and value.ok == "OK";
        """,
        0,
    )

    assert val == 1
    assert "r1" not in r1
    assert "r2" not in r2


def test_eval_incrbyfloat(r: redis.Redis):
    r.set("foo", 0.5)
    val = r.eval(
        """
        local value = redis.call("INCRBYFLOAT", KEYS[1], 2.0);
        return type(value) == "string" and tonumber(value) == 2.5;
        """,
        1,
        "foo",
    )
    assert val == 1


def test_eval_lrange(r: redis.Redis):
    r.rpush("foo", "a", "b")
    val = r.eval(
        """
        local value = redis.call("LRANGE", KEYS[1], 0, -1);
        return type(value) == "table" and value[1] == "a" and value[2] == "b";
        """,
        1,
        "foo",
    )
    assert val == 1


def test_eval_ltrim(r: redis.Redis):
    r.rpush("foo", "a", "b", "c", "d")
    val = r.eval(
        """
        local value = redis.call("LTRIM", KEYS[1], 1, 2);
        return type(value) == "table" and value.ok == "OK";
        """,
        1,
        "foo",
    )
    assert val == 1
    assert r.lrange("foo", 0, -1) == [b"b", b"c"]


def test_eval_lset(r: redis.Redis):
    r.rpush("foo", "a", "b")
    val = r.eval(
        """
        local value = redis.call("LSET", KEYS[1], 0, "z");
        return type(value) == "table" and value.ok == "OK";
        """,
        1,
        "foo",
    )
    assert val == 1
    assert r.lrange("foo", 0, -1) == [b"z", b"b"]


def test_eval_sdiff(r: redis.Redis):
    r.sadd("foo", "a", "b", "c", "f", "e", "d")
    r.sadd("bar", "b")
    val = r.eval(
        """
        local value = redis.call("SDIFF", KEYS[1], KEYS[2]);
        if type(value) ~= "table" then
            return redis.error_reply(type(value) .. ", should be table");
        else
            return value;
        end
        """,
        2,
        "foo",
        "bar",
    )
    # Note: while fakeredis sorts the result when using Lua, this isn't
    # actually part of the redis contract (see
    # https://github.com/antirez/redis/issues/5538), and for Redis 5 we
    # need to sort val to pass the test.
    assert sorted(val) == [b"a", b"c", b"d", b"e", b"f"]


def test_script(r: redis.Redis):
    script = r.register_script("return ARGV[1]")
    result = script(args=[42])
    assert result == b"42"


@testtools.fake_only
def test_lua_log(r, caplog):
    logger = fakeredis._server.LOGGER
    script = """
        redis.log(redis.LOG_DEBUG, "debug")
        redis.log(redis.LOG_VERBOSE, "verbose")
        redis.log(redis.LOG_NOTICE, "notice")
        redis.log(redis.LOG_WARNING, "warning")
    """
    script = r.register_script(script)
    with caplog.at_level("DEBUG"):
        script()
    assert (
        len(
            set(caplog.record_tuples).intersection(
                {
                    (logger.name, logging.DEBUG, "debug"),
                    (logger.name, logging.INFO, "verbose"),
                    (logger.name, logging.INFO, "notice"),
                    (logger.name, logging.WARNING, "warning"),
                }
            )
        )
        == 4
    )


def test_lua_log_no_message(r: redis.Redis):
    script = "redis.log(redis.LOG_DEBUG)"
    script = r.register_script(script)
    with pytest.raises(redis.ResponseError):
        script()


@testtools.fake_only
def test_lua_log_different_types(r, caplog):
    logger = logging.getLogger("fakeredis")
    script = "redis.log(redis.LOG_DEBUG, 'string', 1, true, 3.14, 'string')"
    script = r.register_script(script)
    with caplog.at_level("DEBUG"):
        script()
    assert len(set(caplog.record_tuples).intersection({(logger.name, logging.DEBUG, "string 1 3.14 string")})) == 1


def test_lua_log_wrong_level(r: redis.Redis):
    script = "redis.log(10, 'string')"
    script = r.register_script(script)
    with pytest.raises(redis.ResponseError):
        script()


@testtools.fake_only
def test_lua_log_defined_vars(r, caplog):
    logger = fakeredis._server.LOGGER
    script = """
        local var='string'
        redis.log(redis.LOG_DEBUG, var)
    """
    script = r.register_script(script)
    with caplog.at_level("DEBUG"):
        script()
    assert len(set(caplog.record_tuples).intersection({(logger.name, logging.DEBUG, "string")})) == 1


def test_hscan_cursors_are_bytes(r: redis.Redis):
    r.hset("hkey", "foo", 1)

    result = r.eval(
        """
        local results = redis.call("HSCAN", KEYS[1], "0")
        return results[1]
        """,
        1,
        "hkey",
    )

    assert result == b"0"
    assert isinstance(result, bytes)


@pytest.mark.xfail  # TODO
def test_deleting_while_scan(r: redis.Redis):
    for i in range(100):
        r.set(f"key-{i}", i)

    assert len(r.keys()) == 100

    script = """
        local cursor = 0
        local seen = {}
        repeat
            local result = redis.call('SCAN', cursor)
            for _,key in ipairs(result[2]) do
                seen[#seen+1] = key
                redis.call('DEL', key)
            end
            cursor = tonumber(result[1])
        until cursor == 0
        return seen
    """

    assert len(r.register_script(script)()) == 100
    assert len(r.keys()) == 0


def test_eval_cjson_encode_decode(r: redis.Redis) -> None:
    # Simple encode and decode roundtrip
    lua = """
    local t = {foo = "bar", num = 42}
    local encoded = cjson.encode(t)
    local decoded = cjson.decode(encoded)
    return decoded["foo"] == "bar" and decoded["num"] == 42
    """
    val = r.eval(lua, 0)
    assert val == 1


def test_eval_cjson_no_whitespace(r: redis.Redis) -> None:
    """Cjson library doesn't produce any whitespace on encoding."""
    lua = """
    local t = {foo = "bar", num = 42}
    return cjson.encode(t)
    """
    val = cast(bytes, r.eval(lua, 0))
    # We can't guarantee the order here, so allow both
    assert b'{"foo":"bar","num":42}' == val or b'{"num":42,"foo":"bar"}' == val


def test_eval_cjson_null_decode(r: redis.Redis) -> None:
    # null in JSON becomes cjson.null in Lua
    lua = """
    local json_str = '{"a":null, "b":"val"}'
    local t = cjson.decode(json_str)
    return t["a"] == cjson.null and t["b"] == "val"
    """
    val = r.eval(lua, 0)
    assert val == 1


def test_eval_cjson_null_encode(r: redis.Redis) -> None:
    # Explicitly construct a table with cjson.null and check encoding
    lua = """
    local t = {a = cjson.null, b = "val"}
    return cjson.encode(t)
    """
    val = cast(bytes, r.eval(lua, 0))
    # Order of keys isn't guaranteed, but we can check JSON content
    assert b'"a":null' in val and b'"b":"val"' in val


def test_eval_cjson_nested_structure(r: redis.Redis) -> None:
    lua = """
    local t = {a = {b = {c = 1}}}
    local encoded = cjson.encode(t)
    local decoded = cjson.decode(encoded)
    return decoded["a"]["b"]["c"] == 1
    """
    val = r.eval(lua, 0)
    assert val == 1


def test_eval_cjson_array(r: redis.Redis) -> None:
    lua = """
    local t = {"a", "b", "c"}
    local encoded = cjson.encode(t)
    local decoded = cjson.decode(encoded)
    return decoded[1] == "a" and decoded[2] == "b" and decoded[3] == "c"
    """
    val = r.eval(lua, 0)
    assert val == 1


def test_eval_cjson_dict_array(r: redis.Redis) -> None:
    # lua tables allow a combination of array and dict, cjson should treat this as dict with int keys
    lua = """
    local t = {"a", "b", c=3}
    local encoded = cjson.encode(t)
    local decoded = cjson.decode(encoded)
    return decoded["1"] == "a" and decoded["2"] == "b" and decoded["c"] == 3
    """
    val = r.eval(lua, 0)
    assert val == 1


def test_eval_cjson_mixed(r: redis.Redis) -> None:
    lua = """
    local t = {"a", "b", c={"d", "e", f=3}}
    local encoded = cjson.encode(t)
    local decoded = cjson.decode(encoded)
    return decoded["1"] == "a" and decoded["2"] == "b" and decoded["c"]["1"] == "d" and decoded["c"]["2"] == "e" and decoded["c"]["f"] == 3
    """
    val = r.eval(lua, 0)
    assert val == 1


def test_lock(r: redis.Redis) -> None:
    from redis.lock import Lock

    lock = Lock(r, "my-lock")
    lock.acquire()
    assert lock.locked() is True
    lock.release()


@testtools.fake_only
def test_lua_runtime_reused_across_eval_calls(r: redis.Redis) -> None:
    """LuaRuntime should be cached on FakeServer and reused across eval calls."""
    server = r.connection_pool.connection_kwargs["server"]

    r.eval("return 1", 0)
    runtime_id_1 = id(server._lua_runtime)

    r.eval("return 2", 0)
    runtime_id_2 = id(server._lua_runtime)

    assert runtime_id_1 == runtime_id_2


@pytest.mark.fake
def test_lua_runtime_freed_with_server() -> None:
    """FakeServer (and its cached LuaRuntime) should be garbage collected when destroyed."""
    server = fakeredis.FakeServer()
    r = fakeredis.FakeRedis(server=server)

    r.eval("return 1", 0)
    assert hasattr(server, "_lua_runtime")

    server_ref = weakref.ref(server)
    assert server_ref() is not None

    del r
    del server
    gc.collect()

    assert server_ref() is None


@testtools.fake_only
def test_lua_state_isolated_between_eval_calls(r: redis.Redis) -> None:
    """Lua state should not leak between eval calls when runtime is reused."""
    # First call: set KEYS/ARGV and return them
    result1 = r.eval("return {KEYS[1], ARGV[1]}", 1, "key1", "arg1")
    assert result1 == [b"key1", b"arg1"]

    # Second call: different KEYS/ARGV should be independent
    result2 = r.eval("return {KEYS[1], ARGV[1]}", 1, "key2", "arg2")
    assert result2 == [b"key2", b"arg2"]

    # Third call: no KEYS/ARGV - verify they're reset (nil in Lua = empty table returned)
    result3 = r.eval("return {KEYS[1], ARGV[1]}", 0)
    assert result3 == []  # Lua tables stop at first nil

    # Verify KEYS/ARGV are actually nil, not stale values
    result4 = r.eval("return KEYS[1] == nil and ARGV[1] == nil", 0)
    assert result4 == 1  # true in Lua = 1



"""
Script to create issue for every unsupported command.
To run this:
- Install pygithub: pip install pygithub
- Set environment variable `GITHUB_TOKEN` to a github token with permissions to create issues.
  - Another option is to create `.env` file with `GITHUB_TOKEN`.
"""

import json
import os

import click
import requests
from dotenv import load_dotenv
from github import Github

from scripts.generate_supported_commands_doc import implemented_commands, METADATA, IGNORE_COMMANDS

load_dotenv()  # take environment variables from .env.

THIS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)))

IGNORE_GROUPS = {
    "suggestion",
    "tdigest",
    "scripting",
    "cf",
    "graph",
    "timeseries",
    "connection",
    "server",
    "cms",
    "cluster",
    "search",
    "hash",
}


def commands_groups(all_commands: dict, implemented_set: set) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    implemented, unimplemented = {}, {}
    for cmd in all_commands:
        if cmd.upper() in IGNORE_COMMANDS:
            continue
        group = all_commands[cmd]["group"]
        unimplemented.setdefault(group, [])
        implemented.setdefault(group, [])
        if cmd in implemented_set:
            implemented[group].append(cmd)
        else:
            unimplemented[group].append(cmd)
    return implemented, unimplemented


def download_redis_commands() -> dict:
    cmds = {}
    for item in METADATA:
        full_filename = os.path.join(THIS_DIR, item.local_filename)
        if not os.path.exists(full_filename):
            contents = requests.get(item.url).content
            open(full_filename, "wb").write(contents)
        curr_cmds = json.load(open(full_filename))
        cmds = cmds | {k.lower(): v for k, v in curr_cmds.items()}
    return cmds


def get_unimplemented_and_implemented_commands() -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    """Returns 2 dictionaries, one of unimplemented commands and another of implemented commands"""
    commands = download_redis_commands()
    implemented_commands_set = implemented_commands()
    implemented_dict, unimplemented_dict = commands_groups(commands, implemented_commands_set)
    groups = sorted(implemented_dict.keys(), key=lambda x: len(unimplemented_dict[x]))
    for group in groups:
        unimplemented_count = len(unimplemented_dict[group])
        if unimplemented_count == 0:
            click.secho(f"{group} has all commands implemented", fg="green")
            continue
        total_count = len(implemented_dict.get(group)) + unimplemented_count
        click.secho(f"{group} has {unimplemented_count}/{total_count} unimplemented commands")
    return unimplemented_dict, implemented_dict


class GithubData:
    def __init__(self, dry=False):
        token = os.getenv("GITHUB_TOKEN", None)
        g = Github(token)
        self.dry = dry or (token is None)
        self.gh_repo = g.get_repo("cunla/fakeredis")
        open_issues = self.gh_repo.get_issues(state="open")
        self.issues = {i.title: i.number for i in open_issues}
        gh_labels = self.gh_repo.get_labels()
        self.labels = {label.name for label in gh_labels}

    def create_label(self, name):
        if self.dry:
            click.secho(f'Creating label "{name}"', fg="cyan")
        else:
            self.gh_repo.create_label(name, "f29513")
        self.labels.add(name)

    def create_issue(self, group: str, cmd: str, summary: str):
        link = f"https://redis.io/commands/{cmd.replace(' ', '-')}/"
        title = f"Implement support for `{cmd.upper()}` ({group} command)"
        filename = f"{group}_mixin.py"
        body = f"""Implement support for command `{cmd.upper()}` in {filename}.

{summary}.

Here is the [Official documentation]({link})"""
        labels = [f"{group}-commands", "enhancement", "help wanted"]
        for label in labels:
            if label not in self.labels:
                self.create_label(label)
        if title in self.issues:
            return
        if self.dry:
            print(f'Creating issue with title "{title}" and labels {labels}')
        else:
            self.gh_repo.create_issue(title, body, labels=labels)


def created_issues_for_commands(commands: dict, unimplemented: dict):
    def _handle_commands(gh: GithubData):
        for group in unimplemented:
            if group in IGNORE_GROUPS:
                continue
            if len(unimplemented[group]) == 0:
                continue
            click.secho(f"### Creating issues for {group} commands", fg="green")
            for cmd in unimplemented[group]:
                if cmd.upper() in IGNORE_COMMANDS:
                    continue
                summary = commands[cmd]["summary"]
                gh.create_issue(group, cmd, summary)

    gh = GithubData(dry=True)
    _handle_commands(gh)
    create = input("Press y to create issues, anything else to exit: ")
    if create == "y":
        gh = GithubData()
        _handle_commands(gh)


if __name__ == "__main__":
    commands = download_redis_commands()
    unimplemented_dict, _ = get_unimplemented_and_implemented_commands()
    created_issues_for_commands(commands, unimplemented_dict)



__all__ = [
    "TestConnection",
    "TestHash",
    "TestList",
    "TestServer",
    "TestSet",
    "TestString",
    "TestTransaction",
    "TestZSet",
    "TestZSetNoScores",
    "TestJoint",
]

from .test_connection import TestConnection
from .test_hash import TestHash
from .test_list import TestList
from .test_server import TestServer
from .test_set import TestSet
from .test_string import TestString
from .test_transaction import TestTransaction
from .test_zset import TestZSet, TestZSetNoScores
from .test_hypothesis_joint import TestJoint



"""
Generates a JSON file with the following structure:
{
    "command_name": [
      1. Name
      2. Arity
      3. Flags
      4. First key
      5. Last key
      6. Step
      7. ACL categories (as of Redis 6.0)
      8. Tips (as of Redis 7.0)
      9. Key specifications (as of Redis 7.0)
      10. Subcommands (as of Redis 7.0)
    ]
}
that is used for the `COMMAND` redis command.
"""

import json
import os
from typing import Any, List, Dict

from fakeredis._commands import SUPPORTED_COMMANDS
from scripts.generate_supported_commands_doc import METADATA, download_single_stack_commands

THIS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)))

CATEGORIES = {
    "BF.": "@bloom",
    "CF.": "@cuckoo",
    "CMS.": "@cms",
    "TOPK.": "@topk",
    "TS.": "@timeseries",
    "JSON.": "@json",
    "FT.": "@search",
    "TDIGEST.": "@tdigest",
}


def implemented_commands() -> set:
    res = set(SUPPORTED_COMMANDS.keys())
    if "json.type" not in res:
        raise ValueError("Make sure jsonpath_ng is installed to get accurate documentation")
    if "eval" not in res:
        raise ValueError("Make sure lupa is installed to get accurate documentation")
    return res


def dict_deep_get(d: Dict[Any, Any], *keys, default_value: Any = None) -> Any:
    res = d
    for key in keys:
        if isinstance(res, list) and isinstance(key, int):
            res = res[key]
        else:
            res = res.get(key, None)
        if res is None:
            return default_value
    return default_value if res is None else res


def key_specs_array(cmd_info: Dict[str, Any]) -> List[Any]:
    return []


def get_command_info(cmd_name: str, all_commands: Dict[str, Any]) -> List[Any]:
    """Returns a list
     1 Name //
     2 Arity //
     3 Flags //
     4 First key //
     5 Last key //
     6 Step //
     7 ACL categories (as of Redis 6.0) //
     8 Tips (as of Redis 7.0) //
     9 Key specifications (as of Redis 7.0)
    10 Subcommands (as of Redis 7.0)
    """
    print(f"Command {cmd_name}")
    cmd_info = all_commands[cmd_name]
    first_key = dict_deep_get(cmd_info, "key_specs", 0, "begin_search", "spec", "index", default_value=0)
    last_key = dict_deep_get(cmd_info, "key_specs", -1, "begin_search", "spec", "index", default_value=0)
    step = dict_deep_get(cmd_info, "key_specs", 0, "find_keys", "spec", "keystep", default_value=0)
    tips = []  # todo
    subcommands = [
        get_command_info(cmd, all_commands) for cmd in all_commands if cmd_name != cmd and cmd.startswith(cmd_name)
    ]  # todo
    categories = set(cmd_info.get("acl_categories", []))
    for prefix, category in CATEGORIES.items():
        if cmd_name.startswith(prefix.lower()):
            categories.add(category)
    res = [
        cmd_name.lower().replace(" ", "|"),
        cmd_info.get("arity", -1),
        cmd_info.get("command_flags", []),
        first_key,
        last_key,
        step,
        sorted(categories),
        tips,
        key_specs_array(cmd_info),
        subcommands,
    ]
    return res


if __name__ == "__main__":
    implemented = implemented_commands()
    command_info_dict: Dict[str, List[Any]] = {}
    for cmd_meta in METADATA:
        cmds = download_single_stack_commands(cmd_meta.local_filename, cmd_meta.url, cmd_meta.markdown_commands)
        for cmd in cmds:
            if cmd not in implemented:
                continue
            command_info_dict[cmd] = get_command_info(cmd, cmds)
            subcommand = cmd.split(" ")
            if len(subcommand) > 1:
                command_info_dict.setdefault(
                    subcommand[0],
                    [subcommand[0], -1, [], 0, 0, 0, [], [], [], []],
                )[9].append(command_info_dict[cmd])
            print(command_info_dict[cmd])
    with open(os.path.join(os.path.dirname(__file__), "..", "fakeredis", "commands.json"), "w") as f:
        json.dump(command_info_dict, f)



import pytest
import fakeredis


def test_init_args():
    conn = fakeredis.FakeValkey()
    conn.set("key1", "value1")
    assert conn.get("key1") == b"value1"


@pytest.mark.asyncio
async def test_async_init_kwargs():
    conn = fakeredis.FakeAsyncValkey()
    await conn.set("key2", "value2")
    assert await conn.get("key2") == b"value2"



import threading
import time
from queue import Queue
from typing import Tuple

import pytest
import redis

pytestmark = []
pytestmark.extend(
    [
        pytest.mark.tcp_server,
    ]
)


def test_pubsub(tcp_server_address: Tuple[str, int]):
    def _listen(pubsub, q):
        count = 0
        for message in pubsub.listen():
            q.put(message)
            count += 1
            if count == 4:
                pubsub.close()

    channel = "ch1"
    patterns = ["ch1*", "ch[1]", "ch?"]

    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        pubsub = r.pubsub(ignore_subscribe_messages=True)
        pubsub.subscribe(channel)
        pubsub.psubscribe(*patterns)
        time.sleep(1)

        q = Queue()
        t = threading.Thread(target=_listen, args=(pubsub, q))
        t.start()
        msg = "hello world"
        r.publish(channel, msg)
        t.join()

        msg1 = q.get()
        msg2 = q.get()
        msg3 = q.get()
        msg4 = q.get()

        bpatterns = [pattern.encode() for pattern in patterns]
        bpatterns.append(channel.encode())
        msg = msg.encode()
        assert msg1["data"] == msg
        assert msg1["channel"] in bpatterns
        assert msg2["data"] == msg
        assert msg2["channel"] in bpatterns
        assert msg3["data"] == msg
        assert msg3["channel"] in bpatterns
        assert msg4["data"] == msg
        assert msg4["channel"] in bpatterns



from typing import Tuple

import pytest
import redis


pytestmark = []
pytestmark.extend(
    [
        pytest.mark.tcp_server,
    ]
)


def test_tcp_server_started(tcp_server_address: Tuple[str, int]):
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1]) as r:
        r.set("foo", "bar")
        assert r.get("foo") == b"bar"


def test_tcp_server_connection_reset_error(tcp_server_address: Tuple[str, int]):
    with redis.Redis(*tcp_server_address) as r:
        r.rpush("test", b"foo")

    with redis.Redis(*tcp_server_address) as r:
        assert r.rpop("test") == b"foo"


def test_bulk_string_length(real_server_address: Tuple[str, int], tcp_server_address: Tuple[str, int]):
    """Test that malformed bulk string input is handled correctly."""
    from contextlib import closing
    import socket

    connections = [real_server_address, tcp_server_address]
    for conn in connections:
        host, port = conn[0], conn[1]
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.connect((host, port))
            s.sendall(b"$ 1\ntest")
            data = s.recv(1024).decode()
            assert data == "-ERR unknown command '$', with args beginning with: '1' \r\n", (
                f"Failed for server at {host}:{port}"
            )


def test_tcp_server_started_protocol_3(tcp_server_address: Tuple[str, int]):
    with redis.Redis(host=tcp_server_address[0], port=tcp_server_address[1], protocol=3) as r:
        r.set("foo", "bar")
        assert r.get("foo") == b"bar"



"""
This script generates the markdown files for the supported commands documentation.
"""

import json
import os
from dataclasses import dataclass

import requests
import yaml

from fakeredis._commands import SUPPORTED_COMMANDS

IGNORE_COMMANDS = {
    "QUIT",
    "PUBSUB HELP",
    "FUNCTION HELP",
    "SCRIPT HELP",
    "JSON.DEBUG",
    "JSON.DEBUG HELP",
    "JSON.DEBUG MEMORY",
    "JSON.RESP",
    "XINFO",
    "XINFO HELP",
    "XGROUP",
    "XGROUP HELP",
    "XSETID",
    "ACL HELP",
    "COMMAND HELP",
    "CONFIG HELP",
    "DEBUG",
    "MEMORY HELP",
    "MODULE HELP",
    "CLIENT HELP",
    "PFDEBUG",
    "PFSELFTEST",
    "BITFIELD_RO",
    "OBJECT",
    "OBJECT HELP",
    "OBJECT IDLETIME",
    "OBJECT REFCOUNT",
    "OBJECT FREQ",
    "OBJECT ENCODING",
    "MIGRATE",
    "TOUCH",
}

THIS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)))
markdown_filename_template = "docs/supported-commands/{}.md"


@dataclass
class CommandsMeta:
    local_filename: str
    stack: str
    url: str
    markdown_commands: list[str]


METADATA = [
    CommandsMeta(
        ".commands.json",
        "Redis",
        "https://raw.githubusercontent.com/redis/docs/refs/heads/main/data/commands.json",
        [
            "VADD",
            "VCARD",
            "VDIM",
            "VEMB",
            "VGETATTR",
            "VINFO",
            "VLINKS",
            "VRANDMEMBER",
            "VRANGE",
            "VREM",
            "VSETATTR",
            "VSIM",
            "msetex",
            "hgetex",
            "hsetex",
            "hgetdel",
            "XCFGSET",
        ],
    ),
    CommandsMeta(
        ".json.commands.json",
        "RedisJson",
        "https://raw.githubusercontent.com/redis/docs/refs/heads/main/data/commands_redisjson.json",
        [],
    ),
    CommandsMeta(
        ".ts.commands.json",
        "RedisTimeSeries",
        "https://raw.githubusercontent.com/redis/docs/refs/heads/main/data/commands_redistimeseries.json",
        [],
    ),
    CommandsMeta(
        ".ft.commands.json",
        "RedisSearch",
        "https://raw.githubusercontent.com/redis/docs/refs/heads/main/data/commands_redisearch.json",
        [],
    ),
    CommandsMeta(
        ".bloom.commands.json",
        "RedisBloom",
        "https://raw.githubusercontent.com/redis/docs/refs/heads/main/data/commands_redisbloom.json",
        [],
    ),
]


def download_command_markdown(command: str) -> dict:
    url = f"https://raw.githubusercontent.com/redis/docs/refs/heads/main/content/commands/{command.lower()}.md"
    filename = os.path.join(THIS_DIR, f"{command}.md")
    if not os.path.exists(filename):
        contents = requests.get(url).content
        open(filename, "wb").write(contents)
    # Read markdown file and extract metadata from the top of the file, which is in the format:
    # ---
    # summary: "summary of the command"
    # group: "group of the command"
    # ---
    with open(filename, "r") as f:
        lines = f.readlines()
    metadata = {}
    if lines[0].strip() == "---":
        for ind, line in enumerate(lines[1:], 1):
            if line.strip() == "---":
                break
        metadata = yaml.load("".join(lines[1:ind]), Loader=yaml.SafeLoader)
    return metadata


def download_single_stack_commands(filename, url, markdown_commands: list[str]) -> dict:
    full_filename = os.path.join(THIS_DIR, filename)
    if not os.path.exists(full_filename):
        contents = requests.get(url).content
        open(full_filename, "wb").write(contents)
    curr_cmds = json.load(open(full_filename))
    cmds = {k.lower(): v for k, v in curr_cmds.items()}
    for cmd in markdown_commands:
        if cmd.lower() not in cmds:
            try:
                cmds[cmd.lower()] = download_command_markdown(cmd)
            except Exception as e:
                print(f"Failed to download markdown for command {cmd}: {e}")
    return cmds


def implemented_commands() -> set:
    res = set(SUPPORTED_COMMANDS.keys())
    if "json.type" not in res:
        raise ValueError("Make sure jsonpath_ng is installed to get accurate documentation")
    if "bf.add" not in res:
        raise ValueError("Make sure pybloom-live is installed to get accurate documentation")
    return res


def _commands_groups(commands: dict) -> dict[str, list[str]]:
    groups = {}
    for cmd in commands:
        group = commands[cmd]["group"]
        groups.setdefault(group, []).append(cmd)
    return groups


def generate_redis_commands_markdown_files(redis_commands: dict, fakeredis_commands: set[str], stack: str) -> None:
    groups = _commands_groups(redis_commands)
    for group in groups:
        filename = markdown_filename_template.format(f"{stack}/{group.upper()}")
        with open(filename, "w") as f:
            implemented_in_group = set(groups[group]).intersection(fakeredis_commands)
            implemented_in_group = sorted(implemented_in_group)
            unimplemented_in_group = set(groups[group]) - fakeredis_commands
            unimplemented_in_group = sorted(
                {cmd for cmd in unimplemented_in_group if cmd.upper() not in IGNORE_COMMANDS}
            )
            if len(implemented_in_group) > 0:
                f.write(
                    f"# {stack} `{group}` commands "
                    f"({len(implemented_in_group)}/{len(unimplemented_in_group) + len(implemented_in_group)} "
                    f"implemented)\n\n"
                )
            for cmd in implemented_in_group:
                f.write(f"## [{cmd.upper()}](https://redis.io/commands/{cmd.replace(' ', '-')}/)\n\n")
                f.write(f"{redis_commands[cmd]['summary']}\n\n")
            f.write("\n")

            if len(unimplemented_in_group) > 0:
                f.write(f"## Unsupported {group} commands \n")
                f.write("> To implement support for a command, see [here](/guides/implement-command/) \n\n")
                for cmd in unimplemented_in_group:
                    f.write(
                        f"#### [{cmd.upper()}](https://redis.io/commands/{cmd.replace(' ', '-')}/)"
                        f" <small>(not implemented)</small>\n\n"
                    )
                    f.write(f"{redis_commands[cmd]['summary']}\n\n")
            f.write("\n")


if __name__ == "__main__":
    implemented = implemented_commands()
    non_redis_commands = implemented
    for cmd_meta in METADATA:
        cmds = download_single_stack_commands(cmd_meta.local_filename, cmd_meta.url, cmd_meta.markdown_commands)
        generate_redis_commands_markdown_files(cmds, implemented, cmd_meta.stack)
        non_redis_commands = non_redis_commands - set(cmds.keys())
    print("Commands not in any redis stack:")
    print(non_redis_commands)
